import{e,i as n}from"./core-DdmAo3MM.js";let o=!1;function t(){o=!0}function r(e){if(o)return;const n=document.createElement("div");n.className="py-error",n.textContent=e,n.style.cssText="\n    border: 1px solid red;\n    background: #ffdddd;\n    color: black;\n    font-family: courier, monospace;\n    white-space: pre;\n    overflow-x: auto;\n    padding: 8px;\n    margin-top: 8px;\n  ",document.body.append(n)}e.main.onReady.add((function o(t){e.main.onReady.delete(o);const{stderr:a}=t.io,d=(e,...n)=>(r(e.message||e),a(e,...n));t.io.stderr="py"===t.type?d:n(d),addEventListener("error",(({message:e})=>{e.startsWith("Uncaught PythonError")&&r(e)}))}));export{t as notOnDOM,r as notify};
//# sourceMappingURL=error-DAaZHFER.js.map