mpy-config,mpy-script,py-config,py-script{display:none}.mpy-editor-box,.py-editor-box{padding:.5rem}.mpy-editor-input,.py-editor-input{position:relative}.mpy-editor-box:before,.py-editor-box:before{content:attr(data-env);display:block;font-size:x-small;text-align:end}.mpy-editor-output,.py-editor-output{white-space:pre}.mpy-editor-run-button,.py-editor-run-button{bottom:.5rem;opacity:0;position:absolute;right:.5rem;transition:opacity .25s;z-index:1}.mpy-editor-box:hover .mpy-editor-run-button,.mpy-editor-run-button:disabled,.mpy-editor-run-button:focus,.py-editor-box:hover .py-editor-run-button,.py-editor-run-button:disabled,.py-editor-run-button:focus{opacity:1}@keyframes spinner{to{transform:rotate(1turn)}}.mpy-editor-run-button:disabled>*,.py-editor-run-button:disabled>*{display:none}.mpy-editor-run-button:disabled,.py-editor-run-button:disabled{border-width:0}.mpy-editor-run-button:disabled:before,.py-editor-run-button:disabled:before{animation:spinner .6s linear infinite;background-color:#fff;border:2px solid #aaa;border-radius:50%;border-top-color:#000;box-sizing:border-box;content:"";height:20px;left:100%;margin-left:-26px;margin-top:-23px;position:absolute;top:100%;width:20px}mpy-terminal span,py-terminal span{letter-spacing:0!important}