import{notify as e}from"./error-DAaZHFER.js";import{e as o}from"./core-DdmAo3MM.js";function r(){const e=document.querySelectorAll("script");for(const o of e)s(o.src)}function s(o){/\/pyscript\.net\/latest/.test(o)&&e("Loading scripts from latest is deprecated and will be removed soon. Please use a specific version instead.")}o.main.onReady.add(r),o.main.onWorker.add(r);
//# sourceMappingURL=deprecations-manager-DvN9GcFd.js.map